;; Copyright (c) 2007, Vladimir Sedach

;; All rights reserved.

;; Redistribution and use in source and binary forms, with or without
;; modification, are permitted provided that the following conditions are
;; met: Redistributions of source code must retain the above copyright
;; notice, this list of conditions and the following disclaimer.
;; Redistributions in binary form must reproduce the above copyright
;; notice, this list of conditions and the following disclaimer in the
;; documentation and/or other materials provided with the distribution.

;; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
;; "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
;; LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
;; A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
;; OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
;; SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
;; LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
;; DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
;; THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
;; (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
;; OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

;; Version 1

(defpackage "VAS-RADIX-TREE"
  (:nicknames "RTREE")
  (:use "COMMON-LISP")
  (:export #:make-radix-tree
           #:get-entry
           #:remove-entry))

(in-package "VAS-RADIX-TREE")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; General sequence radix trees

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Basic data structure

(defstruct (radix-tree-node (:conc-name nil))
  (data)
  (leaf)
  (children () :type list)) ;of (edge-string . node)

(defun leaf-p (node)
  (if (leaf node)
      (progn (assert (null (children node)) ()
                     "Radix-tree-node ~A is a leaf but has children!" node) t)
      (assert (null (data node)) ()
              "Radix-tree-node ~A is not a leaf but has data!" node)))

(defmacro edge-string (edge) `(car ,edge))

(defmacro edge-node (edge) `(cdr ,edge))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Interface

(defun make-radix-tree ()
  (make-radix-tree-node :leaf nil))

(defgeneric get-entry (key radix-tree)
  (:documentation "Looks up KEY in the given RADIX-TREE. Returns
either (values <data> t) or (values nil nil <key-matched-upto-index>
<lexicographically-next-largest-suffix>), where (subseq key 0
<key-matched-upto-index>) yields the lexicographically largest
sequence present in the tree smaller than key, and if
<lexicographically-next-largest-suffix> is non-nil, then (concatenate
<type-of-key> (subseq key 0 <key-matched-upto-index>)
<lexicographically-next-largest-suffix>) yields the lexicographically
smallest sequence present in the tree that is larger than key if there
is one."))

(defgeneric remove-entry (key radix-tree)
  (:documentation "Removes the entry with the specified KEY from the
given RADIX-TREE. Returns true if the key was in the tree, and false
otherwise."))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Implementation

;;; Lookup

(defmethod get-entry ((key sequence) (root radix-tree-node))
  (multiple-value-bind (data present-p leaf-node previous-node
                             key-index partial-match-pos partial-match-edge)
      (lookup key root)
    (declare (ignore leaf-node previous-node partial-match-pos))
    (if present-p
        (values data present-p)
        (values nil nil key-index (edge-string partial-match-edge)))))

(defun lookup (key root)
  (lookup% key root root 0))

(defun lookup% (key rtree rtree-prev pos)
  "Returns either (values <data> <present-p> <leaf-node>
<previous-node> <key-index>) or (values nil nil <closest-node>
<closest-node> <key-index> <partial-match-pos> <partial-match-edge>)."
  (if (leaf-p rtree)
      (if (= pos (length key)) ;this is the right leaf node
          (values (data rtree) t rtree rtree-prev pos)
          (values nil nil rtree rtree pos))
      (multiple-value-bind (node pos1 partial-match-edge)
          (lookup-next-node (children rtree) key pos)
        (if node
            (lookup% key node rtree pos1)
            (values nil nil rtree rtree pos pos1 partial-match-edge)))))

(defun lookup-next-node (children key pos)
  "Returns either (values <next-node> <next-pos>) or (values <nil>
<partial-match-pos> <partial-match-edge>) or nil."
  (when children
    (let* ((edge (edge-string (first children)))
           (node (edge-node (first children))))
      (multiple-value-bind (pos1 partial-pos)
          (edge-matches edge key pos)
      (cond (pos1 ;found the next node
             (values node pos1)) 
            (partial-pos ;partial position, key is not in tree
             (values nil partial-pos (first children)))
            (t ;else we need to keep searching
             (lookup-next-node (cdr children) key pos)))))))

(defun edge-matches (edge key pos)
  (let ((edge-length (length edge))
        (key-suffix-length (- (length key) pos)))
    (unless (and (zerop edge-length) (< 0 key-suffix-length))
      (let ((pos1 (mismatch key edge :start1 pos)))
        (if pos1
            (cond ((= edge-length (- pos1 pos)) ;full match, follow this edge
                   pos1)
                  ((zerop (- pos1 pos)) ;no match, ignore this edge
                   nil) 
                  (t ;partial match, key is not in the tree
                   (values nil pos1)))
            (+ edge-length pos))))))

;;; Insert

(defmethod (setf get-entry) (value (key sequence) (root radix-tree-node))
  (insert key root value)
  value)

(defmethod insert ((key sequence) (root radix-tree-node) data)
  (multiple-value-bind (old-data present-p leaf closest-node
                                 key-pos partial-match-pos partial-match-edge)
      (lookup key root)
    (declare (ignore old-data))
    (if present-p
        (setf (data leaf) data)
        (insert-new-key key data closest-node key-pos
                        partial-match-pos partial-match-edge))))

(defun insert-new-key (key data closest-node key-pos partial-match-pos
                       partial-match-edge)
  (let ((new-leaf (make-radix-tree-node :data data :leaf t)))
    (if partial-match-edge
        (split-edge closest-node partial-match-edge key key-pos
                    partial-match-pos new-leaf)
        (insert-new-edge closest-node (subseq key key-pos) new-leaf))))

(defun insert-new-edge (into-node edge-name edge-node)
  (let ((new-edge (cons edge-name edge-node)))
    (if (leaf-p into-node)
        (setf (leaf into-node) nil
              (children into-node)
              (list new-edge
                    (cons (subseq edge-name 0 0)
                          (make-radix-tree-node :leaf t
                                                :data (data into-node))))
              (data into-node) nil)
        (push new-edge (children into-node)))))

(defun split-edge (of-node partial-match-edge key key-pos
                   partial-match-pos new-leaf)
  (destructuring-bind (edge-string . node) partial-match-edge
    (let* ((new-common-substring (subseq key key-pos partial-match-pos))
           (new-edge-string (subseq edge-string (- partial-match-pos key-pos)))
           (new-leaf-suffix (subseq key partial-match-pos))
           (new-subtree
            (make-radix-tree-node
             :leaf nil
             :children (list (cons new-leaf-suffix new-leaf)
                             (cons new-edge-string node)))))
      (remove-edge partial-match-edge of-node)
      (insert-new-edge of-node new-common-substring new-subtree))))

;;; Delete

(defmethod remove-entry ((key sequence) (root radix-tree-node))
  (multiple-value-bind (data present-p leaf predecessor)
      (lookup key root)
    (declare (ignore data))
    (if present-p
        (let ((removed-edge (find leaf (children predecessor) :key #'cdr)))
          (remove-edge removed-edge predecessor)
          (when (and (not (children predecessor)) (not (eq predecessor root)))
            (setf (leaf predecessor) t)
            (remove-entry (subseq key 0 (- (length key)
                                           (length (edge-string removed-edge))))
                          root))
          t)
        nil)))

(defun remove-edge (edge of-node)
  (flet ((sequence-equal (s1 s2)
           (not (mismatch s1 s2))))
    (setf (children of-node) (delete (edge-string edge)
                                     (children of-node)
                                     :key 'car :test #'sequence-equal))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Random tester

(defvar *test-hash* (make-hash-table :test 'equal))
(defvar *test-tree* (make-radix-tree))
(defvar *operations-history* ())

(defun test-cleanup ()
  (setf *test-tree* (make-radix-tree)
        *operations-history* ())
  (clrhash *test-hash*))

(defun run-tests (&key (number 100) (max-key-length 30)
                  (alphabet-size 30) (value-random-limit 100))
  "Randomly runs a battery of operations on the radix tree
implementation. Returns nil if successful, raises an error otherwise."
  (test-cleanup)
  (let ((keys (make-unique-random-keys number max-key-length alphabet-size))
        (inserted-key-bag ())
        (deleted-key-bag ()))
    (labels ((check-tables ()
               (maphash (lambda (key value)
                          (assert (equal value (lookup key *test-tree*)) ()
                                  "Key ~A should map to ~A but maps to ~A!"
                                  key value (lookup key *test-tree*)))
                        *test-hash*))
             (check-deletions ()
               (loop for k in deleted-key-bag do
                     (assert (null (lookup k *test-tree*)) ()
                             "Key ~A is not getting deleted from the tree!" k)))
             (check-bogus ()
               (let ((bogus-keys (make-unique-random-keys
                                  (1+ (random number))
                                  (1+ (random max-key-length))
                                  (1+ (random alphabet-size)))))
                 (loop for key in bogus-keys doing
                       (unless (gethash key *test-hash*)
                         (assert (not (lookup key *test-tree*)))))))
             (check-zombie (rtree)
               (unless (leaf-p rtree)
                 (assert (or (eq rtree *test-tree*) (children rtree)))
                 (map nil (lambda (edge)
                            (check-zombie (edge-node edge)))
                      (children rtree))))
             (verify-consistency ()
               (check-tables)
               (check-deletions)
               (check-bogus)
               (check-zombie *test-tree*)))
      (flet ((insert-key (k &optional (val (random value-random-limit)))
               (setf (gethash k *test-hash*) val)
               (setf (get-entry k *test-tree*) val)
               (pushnew k inserted-key-bag :test 'equal)
               (push (list 'insert-key k) *operations-history*)
               (verify-consistency))
             (delete-key (k)
               (remhash k *test-hash*)
               (remove-entry k *test-tree*)
               (setf inserted-key-bag (delete k inserted-key-bag))
               (pushnew k deleted-key-bag :test 'equal)
               (push (list 'delete-key k) *operations-history*)
               (verify-consistency))
             (get-random-inserted-key ()
               (elt inserted-key-bag (random (length inserted-key-bag)))))
        (loop for k in keys doing
              (let ((state (random 2)))
                (cond ((= 0 state)
                       (insert-key k))
                      ((and (= 1 state) inserted-key-bag)
                       (delete-key (get-random-inserted-key))))))))))

(defun make-unique-random-keys (number key-length alphabet-size)
  (let ((keys ()))
    (loop repeat number do
          (let ((key (make-array (list (random key-length)))))
            (dotimes (i (length key))
              (setf (aref key i) (random alphabet-size)))
            (pushnew (coerce key 'list) keys :test 'equal)))
    keys))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Binary radix trees

;;; TODO